/* Use this Code with ESP32 WROOM-DA Module
*/
#include <BluetoothSerial.h>   // Library for classic Bluetooth
#include <BLEDevice.h>         // Library for BLE Bluetooth
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <esp_task_wdt.h>

// UART configuration (change as needed)
#define RX1_PIN 16   // RX pin for UART2
#define TX1_PIN 17   // TX pin for UART2

// Parametrized device names
const char* esp32ServerName = "ESP32_Server"; // Name of the ESP32 device as a server
const char* bleName = "ESP32_BT_BLE"; // Name of the ESP32 BT for Android or BLE for iOS

// LED pin
const int ledPin = 2;  // Adjust if the LED is on a different pin

// BLE UUIDs for the Nordic UART Service (NUS)
#define SERVICE_UUID           "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_RX "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"

// Watchdog timeout in seconds
#define WDT_TIMEOUT 60

BLEServer* pServer = NULL;
BLECharacteristic* pTxCharacteristic = NULL;
BLECharacteristic* pRxCharacteristic = NULL;
bool bleDeviceConnected = false;
bool oldBleDeviceConnected = false;

bool btDeviceConnected = false;

// Bluetooth for Android (Server)
BluetoothSerial SerialBTServer;

unsigned long lastUartDataTime = 0;
const unsigned long uartTimeout = 5000; // 5 seconds timeout for UART
unsigned long uartBaudrate = 115200;


void blinkLED(uint32_t millis)
{
    //delay(millis);
    digitalWrite(ledPin, HIGH);
    delay(millis);
    digitalWrite(ledPin, LOW);
}

// Callback class for BLE
class MyServerCallbacks : public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      bleDeviceConnected = true;
      Serial.println("BLE device connected.");
    };

    void onDisconnect(BLEServer* pServer) {
      bleDeviceConnected = false;
      Serial.println("BLE device disconnected.");
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic* pCharacteristic) override {
        std::string value = pCharacteristic->getValue();
        if (value.length() > 0) {
            Serial1.write((uint8_t*)value.c_str(), value.length());
            blinkLED(1);
        }
    }
};

// Callback function for classic Bluetooth (Server for Android)
void btServerCallback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if (event == ESP_SPP_SRV_OPEN_EVT) {
    btDeviceConnected = true;
    Serial.println("Classic Bluetooth device connected (Android).");
  } else if (event == ESP_SPP_CLOSE_EVT) {
    btDeviceConnected = false;
    Serial.println("Classic Bluetooth device disconnected (Android).");
  }
}

// Function for the UART task
void uartTask(void *pvParameters) {
  while (1) {
    // Receive data from UART and forward to BLE (iOS) or Bluetooth device (Android)
    while (Serial1.available()) {
      size_t len = Serial1.available();
      uint8_t buffer[len];
      Serial1.readBytes(buffer, len);
      
      if (bleDeviceConnected) {
        pTxCharacteristic->setValue(buffer, len);
        pTxCharacteristic->notify(); // Send data to connected iOS device
        blinkLED(1);
      } else if (btDeviceConnected) {
        SerialBTServer.write(buffer, len);  // Send data to connected Android device
        blinkLED(1);
      } else {
        Serial.println("No BLE or BT device connected. Data not sent.");
        blinkLED(1000);
      }
      lastUartDataTime = millis();
    }
    vTaskDelay(1 / portTICK_PERIOD_MS);
  }
}

// Callback function called when data is received via UART1
void onUartReceive() {
  // Receive data from UART and send to BLE or Bluetooth
  while (Serial1.available()) {
    
    if (bleDeviceConnected) {
      size_t len = Serial1.available();
      uint8_t buffer[len];
      Serial1.readBytes(buffer, len);
      pTxCharacteristic->setValue(buffer, len);
      //Serial.write(buffer, len); // Output to the serial console
      pTxCharacteristic->notify();
      blinkLED(1);
      
    } else if (btDeviceConnected) {
      uint8_t receivedByte = Serial1.read();
      SerialBTServer.write(receivedByte);  // Send data to connected Android device
      //Serial.write(receivedByte);  // Output to the serial console
    } else {
      Serial.println("No BLE or BT device connected. Data not sent.");
      blinkLED(1000);
    }
  }
  blinkLED(1);
}

// Callback function triggered when data is received via classic Bluetooth
void onBluetoothData(const uint8_t* buffer, size_t size) {
  // Process the received data and write it to UART
  for (size_t i = 0; i < size; i++) {
    Serial1.write(buffer[i]);  // Write the data to UART
    //Serial.write(buffer[i]);   // Optionally, write the data to the serial console
  }
  blinkLED(1);  // Blink the LED to indicate data reception
}

void setup() {
  Serial.begin(115200);

  esp_task_wdt_init(WDT_TIMEOUT, true);  // WDT_TIMEOUT seconds, automatically reset the system on timeout
  esp_task_wdt_add(NULL);  // Add the current task to the watchdog

  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, LOW);

  // Initialize UART1 with 115200 baud rate
  Serial1.begin(uartBaudrate, SERIAL_8N1, RX1_PIN, TX1_PIN);

  // Add receive callback for UART1
  //Serial1.onReceive(onUartReceive);  // Register the callback function.

  // Start classic Bluetooth as a server for Android devices and register callback
  if (SerialBTServer.begin(esp32ServerName)) {
    SerialBTServer.register_callback(btServerCallback);
    SerialBTServer.onData(onBluetoothData);  // Register the onData callback
    Serial.println("Classic Bluetooth server started. Waiting for connection to Android device...");
  } else {
    Serial.println("Error starting classic Bluetooth (Server for Android).");
  }

  // BLE initialization for IOS devices
  BLEDevice::init(bleName); 
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  BLEService* pService = pServer->createService(SERVICE_UUID);
  pTxCharacteristic = pService->createCharacteristic(
                        CHARACTERISTIC_UUID_TX,
                        BLECharacteristic::PROPERTY_READ |
                        BLECharacteristic::PROPERTY_NOTIFY
                      );
  pTxCharacteristic->addDescriptor(new BLE2902());

  pRxCharacteristic = pService->createCharacteristic(
                        CHARACTERISTIC_UUID_RX,
                        BLECharacteristic::PROPERTY_WRITE
                      );
  pRxCharacteristic->setCallbacks(new MyCallbacks());

  pService->start();

  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(true);
  pAdvertising->setMinPreferred(0x06);  
  pAdvertising->setMinPreferred(0x12);
  pAdvertising->start();
  Serial.println("BLE started and advertising sent.");

  // Start the UART task only if code line 'Serial1.onReceive(onUartReceive);' is deactivated 
   xTaskCreatePinnedToCore(uartTask, "UART Task", 8192, NULL, 1, NULL, 1);
}

void loop() {
  if (bleDeviceConnected && !oldBleDeviceConnected) {
    oldBleDeviceConnected = bleDeviceConnected;
  }

  if (!bleDeviceConnected && oldBleDeviceConnected) {
    oldBleDeviceConnected = bleDeviceConnected;
    pServer->startAdvertising();
    Serial.println("Advertising restarted.");
  }

  esp_task_wdt_reset();  // Watchdog reset
  delay(10);
}
